"""
Generated by the output of the 'protocol' command of signald
Version: '0.12.0+git2021-01-30rce7ef396.5'
"""
from dataclasses import dataclass, field
import typing

from pysignald_async.util import JSONProtocol, nested_dataclass, locals_to_request


@nested_dataclass
class JsonMessageEnvelopev1:
    username: str = None
    uuid: str = None
    source: "JsonAddressv1" = None
    sourceDevice: int = None
    type: str = None
    relay: str = None
    timestamp: int = None
    timestampISO: str = None
    serverTimestamp: int = None
    serverDeliveredTimestamp: int = None
    hasLegacyMessage: bool = None
    hasContent: bool = None
    isUnidentifiedSender: bool = None
    dataMessage: "JsonDataMessagev1" = None
    syncMessage: "JsonSyncMessagev1" = None
    callMessage: "JsonCallMessagev0" = None
    receipt: "JsonReceiptMessagev0" = None
    typing: "JsonTypingMessagev0" = None


@nested_dataclass
class SendResponsev1:
    results: typing.List["JsonSendMessageResultv1"] = field(default_factory=list)
    timestamp: int = None


@nested_dataclass
class JsonVersionMessagev1:
    name: str = None
    version: str = None
    branch: str = None
    commit: str = None


@nested_dataclass
class JsonGroupV2Infov1:
    id: str = None
    revision: int = None
    title: str = None
    timer: int = None
    members: typing.List["JsonAddressv1"] = field(default_factory=list)
    pendingMembers: typing.List["JsonAddressv1"] = field(default_factory=list)
    requestingMembers: typing.List["JsonAddressv1"] = field(default_factory=list)
    inviteLink: str = None


@nested_dataclass
class LinkedDevicesv1:
    devices: typing.List["DeviceInfov0"] = field(default_factory=list)


@nested_dataclass
class JsonGroupJoinInfov1:
    groupID: str = None
    title: str = None
    memberCount: int = None
    addFromInviteLink: int = None
    revision: int = None
    pendingAdminApproval: bool = None


@nested_dataclass
class GroupInfov1:
    """
    A generic type that is used when the group version is not known
    """

    v1: "JsonGroupInfov1" = None
    v2: "JsonGroupV2Infov1" = None


@nested_dataclass
class SetProfilev1:
    account: str = None
    name: str = None
    avatarFile: str = None


@nested_dataclass
class JsonAddressv1:
    number: str = None
    uuid: str = None
    relay: str = None


@nested_dataclass
class Profilev1:
    """
    Information about a Signal user
    """

    name: str = None
    avatar: str = None
    address: "JsonAddressv1" = None
    capabilities: "Capabilitiesv1" = None
    color: str = None
    profile_name: str = None
    inbox_position: int = None
    expiration_time: int = None


@nested_dataclass
class GroupListv1:
    groups: typing.List["JsonGroupV2Infov1"] = field(default_factory=list)
    legacyGroups: typing.List["JsonGroupInfov1"] = field(default_factory=list)


@nested_dataclass
class ProfileListv1:
    profiles: typing.List["Profilev1"] = field(default_factory=list)


@nested_dataclass
class JsonDataMessagev1:
    timestamp: int = None
    attachments: typing.List["JsonAttachmentv0"] = field(default_factory=list)
    body: str = None
    group: "JsonGroupInfov1" = None
    groupV2: "JsonGroupV2Infov1" = None
    endSession: bool = None
    expiresInSeconds: int = None
    profileKeyUpdate: bool = None
    quote: "JsonQuotev1" = None
    contacts: typing.List["SharedContactv0"] = field(default_factory=list)
    previews: typing.List["JsonPreviewv0"] = field(default_factory=list)
    sticker: "JsonStickerv0" = None
    viewOnce: bool = None
    reaction: "JsonReactionv1" = None
    remoteDelete: "RemoteDeletev0" = None
    mentions: typing.List["JsonMentionv1"] = field(default_factory=list)


@nested_dataclass
class JsonSyncMessagev1:
    sent: "JsonSentTranscriptMessagev1" = None
    contacts: "JsonAttachmentv0" = None
    contactsComplete: bool = None
    groups: "JsonAttachmentv0" = None
    blockedList: "JsonBlockedListMessagev1" = None
    request: str = None
    readMessages: typing.List["JsonReadMessagev1"] = field(default_factory=list)
    viewOnceOpen: "JsonViewOnceOpenMessagev1" = None
    verified: "JsonVerifiedMessagev1" = None
    configuration: "ConfigurationMessagev0" = None
    stickerPackOperations: typing.List["JsonStickerPackOperationMessagev0"] = field(
        default_factory=list
    )
    fetchType: str = None
    messageRequestResponse: "JsonMessageRequestResponseMessagev1" = None


@nested_dataclass
class JsonQuotev1:
    """
    A quote is a reply to a previous message. ID is the sent time of the message being replied to
    """

    id: int = None
    author: "JsonAddressv1" = None
    text: str = None
    attachments: typing.List["JsonQuotedAttachmentv0"] = field(default_factory=list)
    mentions: typing.List["JsonMentionv1"] = field(default_factory=list)


@nested_dataclass
class JsonMentionv1:
    uuid: str = None
    start: int = None
    length: int = None


@nested_dataclass
class JsonSendMessageResultv1:
    address: "JsonAddressv1" = None
    success: "Successv0" = None
    networkFailure: bool = None
    unregisteredFailure: bool = None
    identityFailure: str = None


@nested_dataclass
class JsonReactionv1:
    emoji: str = None
    remove: bool = None
    targetAuthor: "JsonAddressv1" = None
    targetSentTimestamp: int = None


@nested_dataclass
class JsonGroupInfov1:
    groupId: str = None
    members: typing.List["JsonAddressv1"] = field(default_factory=list)
    name: str = None
    type: str = None
    avatarId: int = None


@nested_dataclass
class Capabilitiesv1:
    gv2: bool = None
    storage: bool = None
    gv1_migration: bool = None


@nested_dataclass
class JsonSentTranscriptMessagev1:
    destination: "JsonAddressv1" = None
    timestamp: int = None
    expirationStartTimestamp: int = None
    message: "JsonDataMessagev1" = None
    unidentifiedStatus: dict = None
    isRecipientUpdate: bool = None


@nested_dataclass
class JsonBlockedListMessagev1:
    addresses: typing.List["JsonAddressv1"] = field(default_factory=list)
    groupIds: typing.List[str] = field(default_factory=list)


@nested_dataclass
class JsonReadMessagev1:
    sender: "JsonAddressv1" = None
    timestamp: int = None


@nested_dataclass
class JsonViewOnceOpenMessagev1:
    sender: "JsonAddressv1" = None
    timestamp: int = None


@nested_dataclass
class JsonVerifiedMessagev1:
    destination: "JsonAddressv1" = None
    identityKey: str = None
    verified: str = None
    timestamp: int = None


@nested_dataclass
class JsonAccountListv0:
    accounts: typing.List["JsonAccountv0"] = field(default_factory=list)


@nested_dataclass
class JsonCallMessagev0:
    offerMessage: "OfferMessagev0" = None
    answerMessage: "AnswerMessagev0" = None
    busyMessage: "BusyMessagev0" = None
    hangupMessage: "HangupMessagev0" = None
    iceUpdateMessages: typing.List["IceUpdateMessagev0"] = field(default_factory=list)
    destinationDeviceId: int = None
    isMultiRing: bool = None


@nested_dataclass
class JsonReceiptMessagev0:
    type: str = None
    timestamps: typing.List[int] = field(default_factory=list)
    when: int = None


@nested_dataclass
class JsonTypingMessagev0:
    action: str = None
    timestamp: int = None
    groupId: str = None


@nested_dataclass
class JsonAccountv0:
    deviceId: int = None
    username: str = None
    filename: str = None
    uuid: str = None
    registered: bool = None
    has_keys: bool = None
    subscribed: bool = None


@nested_dataclass
class JsonAttachmentv0:
    contentType: str = None
    id: str = None
    size: int = None
    storedFilename: str = None
    filename: str = None
    customFilename: str = None
    caption: str = None
    width: int = None
    height: int = None
    voiceNote: bool = None
    key: str = None
    digest: str = None
    blurhash: str = None


@nested_dataclass
class DeviceInfov0:
    id: int = None
    name: str = None
    created: int = None
    lastSeen: int = None


@nested_dataclass
class SharedContactv0:
    name: "Namev0" = None
    avatar: "Optionalv0" = None
    phone: "Optionalv0" = None
    email: "Optionalv0" = None
    address: "Optionalv0" = None
    organization: "Optionalv0" = None


@nested_dataclass
class JsonPreviewv0:
    url: str = None
    title: str = None
    attachment: "JsonAttachmentv0" = None


@nested_dataclass
class JsonStickerv0:
    packID: str = None
    packKey: str = None
    stickerID: int = None
    attachment: "JsonAttachmentv0" = None


@nested_dataclass
class RemoteDeletev0:
    targetSentTimestamp: int = None


@nested_dataclass
class ConfigurationMessagev0:
    readReceipts: "Optionalv0" = None
    unidentifiedDeliveryIndicators: "Optionalv0" = None
    typingIndicators: "Optionalv0" = None
    linkPreviews: "Optionalv0" = None


@nested_dataclass
class JsonStickerPackOperationMessagev0:
    packID: str = None
    packKey: str = None
    type: str = None


@nested_dataclass
class OfferMessagev0:
    id: int = None
    sdp: str = None
    type: "Typev0" = None
    opaque: str = None


@nested_dataclass
class AnswerMessagev0:
    id: int = None
    sdp: str = None
    opaque: str = None


@nested_dataclass
class BusyMessagev0:
    id: int = None


@nested_dataclass
class HangupMessagev0:
    id: int = None
    type: "Typev0" = None
    deviceId: int = None
    legacy: bool = None


@nested_dataclass
class IceUpdateMessagev0:
    id: int = None
    opaque: str = None
    sdp: str = None


@nested_dataclass
class JsonQuotedAttachmentv0:
    contentType: str = None
    fileName: str = None
    thumbnail: "JsonAttachmentv0" = None


@nested_dataclass
class Successv0:
    unidentified: bool = None
    needsSync: bool = None
    duration: int = None


@nested_dataclass
class Namev0:
    display: "Optionalv0" = None
    given: "Optionalv0" = None
    family: "Optionalv0" = None
    prefix: "Optionalv0" = None
    suffix: "Optionalv0" = None
    middle: "Optionalv0" = None


@nested_dataclass
class Optionalv0:
    present: bool = None


@nested_dataclass
class Typev0:
    pass


@nested_dataclass
class LinkedDevicesv1alpha1:
    devices: typing.List["DeviceInfov0"] = field(default_factory=list)


@nested_dataclass
class JsonGroupJoinInfov1alpha1:
    groupID: str = None
    title: str = None
    memberCount: int = None
    addFromInviteLink: int = None
    revision: int = None
    pendingAdminApproval: bool = None


@nested_dataclass
class JsonGroupV2Infov1alpha1:
    id: str = None
    revision: int = None
    title: str = None
    timer: int = None
    members: typing.List["JsonAddressv1"] = field(default_factory=list)
    pendingMembers: typing.List["JsonAddressv1"] = field(default_factory=list)
    requestingMembers: typing.List["JsonAddressv1"] = field(default_factory=list)
    inviteLink: str = None


@nested_dataclass
class GroupInfov1alpha2:
    """
    A generic type that is used when the group version is not known
    """

    v1: "JsonGroupInfov1" = None
    v2: "JsonGroupV2Infov1" = None


class SignaldGeneratedAPI(JSONProtocol):
    async def send(
        self,
        username: str = None,
        recipientAddress: "JsonAddressv1" = None,
        recipientGroupId: str = None,
        messageBody: str = None,
        attachments: typing.List["JsonAttachmentv0"] = None,
        quote: "JsonQuotev1" = None,
        timestamp: int = None,
        mentions: typing.List["JsonMentionv1"] = None,
    ) -> SendResponsev1:
        """


        username: Example: "+12024561414"
        recipientAddress:
        recipientGroupId: Example: "EdSqI90cS0UomDpgUXOlCoObWvQOXlH5G3Z2d3f4ayE="
        messageBody: Example: "hello"
        attachments:
        quote:
        timestamp:
        mentions:
        """

        return SendResponsev1(
            **await self.get_response(
                {"type": "send", "version": "v1", **locals_to_request(locals())}
            )
        )

    async def react(
        self,
        username: str = None,
        recipientAddress: "JsonAddressv1" = None,
        recipientGroupId: str = None,
        reaction: "JsonReactionv1" = None,
        timestamp: int = None,
    ) -> SendResponsev1:
        """
        react to a previous message

        username: Example: "+12024561414"
        recipientAddress:
        recipientGroupId: Example: "EdSqI90cS0UomDpgUXOlCoObWvQOXlH5G3Z2d3f4ayE="
        reaction:
        timestamp:
        """

        return SendResponsev1(
            **await self.get_response(
                {"type": "react", "version": "v1", **locals_to_request(locals())}
            )
        )

    async def version(
        self,
    ) -> JsonVersionMessagev1:
        """"""

        return JsonVersionMessagev1(
            **await self.get_response(
                {"type": "version", "version": "v1", **locals_to_request(locals())}
            )
        )

    async def accept_invitation(
        self,
        account: str = None,
        groupID: str = None,
    ) -> JsonGroupV2Infov1:
        """
        Accept a v2 group invitation. Note that you must have a profile name set to join groups.

        account: The account to interact with. Example: "+12024561414"
        groupID: Example: "EdSqI90cS0UomDpgUXOlCoObWvQOXlH5G3Z2d3f4ayE="
        """

        return JsonGroupV2Infov1(
            **await self.get_response(
                {
                    "type": "accept_invitation",
                    "version": "v1",
                    **locals_to_request(locals()),
                }
            )
        )

    async def approve_membership(
        self,
        account: str = None,
        groupID: str = None,
        members: typing.List["JsonAddressv1"] = None,
    ) -> JsonGroupV2Infov1:
        """
        approve a request to join a group

        account: The account to interact with. Example: "+12024561414"
        groupID: Example: "EdSqI90cS0UomDpgUXOlCoObWvQOXlH5G3Z2d3f4ayE="
        members: list of requesting members to approve.
        """

        return JsonGroupV2Infov1(
            **await self.get_response(
                {
                    "type": "approve_membership",
                    "version": "v1",
                    **locals_to_request(locals()),
                }
            )
        )

    async def get_group(
        self,
        account: str = None,
        groupID: str = None,
        revision: int = None,
    ) -> JsonGroupV2Infov1:
        """
        Query the server for the latest state of a known group

        account: The account to interact with. Example: "+12024561414"
        groupID: Example: "EdSqI90cS0UomDpgUXOlCoObWvQOXlH5G3Z2d3f4ayE="
        revision: the latest known revision, default value (-1) forces fetch from server.
        """

        return JsonGroupV2Infov1(
            **await self.get_response(
                {"type": "get_group", "version": "v1", **locals_to_request(locals())}
            )
        )

    async def get_linked_devices(
        self,
        account: str = None,
    ) -> LinkedDevicesv1:
        """
        list all linked devices on a Signal account

        account: The account to interact with. Example: "+12024561414"
        """

        return LinkedDevicesv1(
            **await self.get_response(
                {
                    "type": "get_linked_devices",
                    "version": "v1",
                    **locals_to_request(locals()),
                }
            )
        )

    async def join_group(
        self,
        account: str = None,
        uri: str = None,
    ) -> JsonGroupJoinInfov1:
        """
        Join a group using the a signal.group URL. Note that you must have a profile name set to join groups.

        account: The account to interact with. Example: "+12024561414"
        uri: The signal.group URL. Example: "https://signal.group/#CjQKINH_GZhXhfifTcnBkaKTNRxW-hHKnGSq-cJNyPVqHRp8EhDUB7zjKNEl0NaULhsqJCX3"
        """

        return JsonGroupJoinInfov1(
            **await self.get_response(
                {"type": "join_group", "version": "v1", **locals_to_request(locals())}
            )
        )

    async def protocol(
        self,
    ):
        """"""

        await self.get_response(
            {"type": "protocol", "version": "v1", **locals_to_request(locals())}
        )

    async def remove_linked_device(
        self,
        account: str = None,
        deviceId: int = None,
    ):
        """
        Remove a linked device from the Signal account. Only allowed when the local device id is 1

        account: The account to interact with. Example: "+12024561414"
        deviceId: the ID of the device to unlink. Example: 3
        """

        await self.get_response(
            {
                "type": "remove_linked_device",
                "version": "v1",
                **locals_to_request(locals()),
            }
        )

    async def update_group(
        self,
        account: str = None,
        groupID: str = None,
        title: str = None,
        avatar: str = None,
        addMembers: typing.List["JsonAddressv1"] = None,
        removeMembers: typing.List["JsonAddressv1"] = None,
    ) -> GroupInfov1:
        """
        modify a group

        account: The identifier of the account to interact with. Example: "+12024561414"
        groupID: Example: "EdSqI90cS0UomDpgUXOlCoObWvQOXlH5G3Z2d3f4ayE="
        title: Example: "Parkdale Run Club"
        avatar: Example: "/tmp/image.jpg"
        addMembers:
        removeMembers:
        """

        return GroupInfov1(
            **await self.get_response(
                {"type": "update_group", "version": "v1", **locals_to_request(locals())}
            )
        )

    async def set_profile(
        self,
        account: str = None,
        name: str = None,
        avatarFile: str = None,
    ):
        """


        account: The phone number of the account to use. Example: "+12024561414"
        name: New profile name. Set to empty string for no profile name. Example: "signald user"
        avatarFile: Path to new profile avatar file, if the avatar should be updated. Example: "/tmp/image.jpg"
        """

        await self.get_response(
            {"type": "set_profile", "version": "v1", **locals_to_request(locals())}
        )

    async def resolve_address(
        self,
        account: str = None,
        partial: "JsonAddressv1" = None,
    ) -> JsonAddressv1:
        """
        Resolve a partial JsonAddress with only a number or UUID to one with both. Anywhere that signald accepts a JsonAddress will except a partial, this is a convenience function for client authors, mostly because signald doesn't resolve all the partials it returns

        account: The signal account to use. Example: "+12024561414"
        partial: The partial address, missing fields.
        """

        return JsonAddressv1(
            **await self.get_response(
                {
                    "type": "resolve_address",
                    "version": "v1",
                    **locals_to_request(locals()),
                }
            )
        )

    async def mark_read(
        self,
        account: str = None,
        to: "JsonAddressv1" = None,
        timestamps: typing.List[int] = None,
    ):
        """


        account: The account to interact with. Example: "+12024561414"
        to: The address that sent the message being marked as read.
        timestamps: List of messages to mark as read. Example: 1612005182278
        """

        await self.get_response(
            {"type": "mark_read", "version": "v1", **locals_to_request(locals())}
        )

    async def get_profile(
        self,
        account: str = None,
        async_: bool = None,
        address: "JsonAddressv1" = None,
    ) -> Profilev1:
        """
        Get all information available about a user

        account: the signald account to use.
        async_: return results from local store immediately, refreshing from server if needed. If false (default), block until all pending profiles have been retrieved..
        address: the address to look up.
        """

        return Profilev1(
            **await self.get_response(
                {"type": "get_profile", "version": "v1", **locals_to_request(locals())}
            )
        )

    async def list_groups(
        self,
        account: str = None,
    ) -> GroupListv1:
        """


        account:
        """

        return GroupListv1(
            **await self.get_response(
                {"type": "list_groups", "version": "v1", **locals_to_request(locals())}
            )
        )

    async def list_contacts(
        self,
        account: str = None,
        async_: bool = None,
    ) -> ProfileListv1:
        """


        account:
        async_: return results from local store immediately, refreshing from server if needed. If false (default), block until all pending profiles have been retrieved..
        """

        return ProfileListv1(
            **await self.get_response(
                {
                    "type": "list_contacts",
                    "version": "v1",
                    **locals_to_request(locals()),
                }
            )
        )



from dataclasses import dataclass, is_dataclass, fields
import asyncio
import json
import logging
import random
import typing
import copy
import re

from pysignald_async.error import raise_error


@dataclass
class Handler:
    """
    Allows to await for a specific message sent by JSONProtocol.
    This should not be used directly, but rather using
    JSONProtocol.get_future_for(validator).

    validator: a Callable that will receive the response and return True
        if the response is the one it was waiting for
    callback: a Callable that will be called with the response as argument
    """

    validator: typing.Callable
    callback: typing.Callable

    def validate(self, response):
        return self.validator(response)


class JSONProtocol(asyncio.Protocol):
    def __init__(self, on_con_lost: asyncio.Future = None):
        """
        on_con_lost: asyncio.Future: its result will be set to True once
            the connection is lost
        """
        self._buffer = bytearray()
        self.logger = logging.getLogger(self.__class__.__name__)
        self.transport = None

        if on_con_lost is None:
            on_con_lost = asyncio.get_running_loop().create_future()

        self.on_con_lost = on_con_lost
        self.callbacks = {}
        self.specific_handlers = []

    def connection_made(self, transport: asyncio.Transport):
        self.logger.info("Connection established")
        self.transport = transport

    def connection_lost(self, exc: typing.Union[Exception, None]):
        self.logger.info(f"Connection lost: {exc}")
        self.transport = None
        try:
            self.on_con_lost.set_result(True)
        except asyncio.InvalidStateError:
            pass

    def data_received(self, data: bytes):
        """
        Handle data received through the UNIX socket, convert it to a python
        dict and dispatch it to the relevant callback or handler.

        Callbacks can either:
        - wait for a specific payload id, for this you should use the coroutine
          JSONProtocol.get_response
        - wait for a more generic matching payload, for this you should use the
          coroutine JSONProtocol.get_future_for

        In case no callbacks match the payload, it is sent to the method
        JSONProtocol.handle_{payload.type}.
        """
        buffer = self._buffer
        buffer += data
        if not buffer.endswith(b"\n"):
            # print(data)
            return
        # data = self._buffer
        data = buffer
        self._buffer = bytearray()

        for line in data.decode("utf-8").split("\n"):
            if not line:
                continue  # Empty lines are sometimes sent apparently…

            payload = json.loads(line)
            self.logger.debug(
                "Received payload:"
                + "\n"
                + json.dumps(payload, indent=4, sort_keys=True)
            )

            for handler in self.specific_handlers:
                if handler.validate(payload):
                    self.logger.debug(f"Found a specific handler")
                    self.specific_handlers.remove(handler)
                    handler.callback(payload)
                    return

            id_ = payload.get("id")
            if id_ is None:
                type_ = payload.get("type")
                try:
                    handler = getattr(self, f"handle_{type_}")
                except AttributeError:
                    self.logger.info(f"No method to handle {type_}, ignoring")
                    return
                handler(payload)
            else:
                callback = self.callbacks.pop(id_, None)
                if callback is None:
                    self.logger.warning(
                        f"Received payload with id but no callbacks were"
                        "registered for it, ignoring"
                    )
                else:
                    callback(payload)

    def send_request(self, payload: dict, id_=None, callback=None):
        """
        Send a JSON payload to the UNIX socket.

        Arguments:
        payload: dict
        id_: identifier of the request. If not specified, a random string is used.
        callback: a Callable that will be called with the response payload
        """
        if self.transport is None:
            self.logger.warning("No transport, cannot send payload")
            return

        if id_ is None:
            id_ = random_id()
        payload["id"] = id_

        if callback is not None:
            self.callbacks[id_] = callback

        self.logger.debug(
            "Send payload" + "\n" + json.dumps(payload, indent=4, sort_keys=True)
        )
        self.transport.write(json.dumps(payload).encode("utf8") + b"\n")

    def get_future_for(self, validator: typing.Callable) -> asyncio.Future:
        """
        Listens to a specific payload with various criteria. Once they are matched,
        set the result of the Future returned by this function to the payload.

        Criteria are matched with the Handler class.
        """
        loop = asyncio.get_running_loop()
        future = loop.create_future()
        handler = Handler(
            validator=validator,
            callback=lambda res: future.set_result(res),
        )
        # monkey patching Future to have a reference to the handler in case we finally
        # dont want to wait for the future to resolve
        future.handler = handler
        self.specific_handlers.append(handler)
        return future

    async def get_response(self, payload: dict) -> dict:
        """
        Coroutine to await the response to a specific payload.
        Can raise exceptions in case the response is an error.
        """
        loop = asyncio.get_running_loop()
        future = loop.create_future()
        self.send_request(
            payload=payload, callback=lambda payload: future.set_result(payload)
        )
        response = await future
        raise_error(response)
        return response.get("data", dict())


def random_id(length=20):
    return "".join(
        random.choice("abcdefghijklmnopqrstuvwxyz0123456789") for _ in range(length)
    )


# Vastly inspired from
# https://www.geeksforgeeks.org/creating-nested-dataclass-objects-in-python/
def nested_dataclass(*args, **kwargs):
    def wrapper(check_class):
        # needed here because of circular imports and problem with type
        # annotations
        import pysignald_async.generated as signald_api

        # passing class to investigate
        check_class = dataclass(check_class, **kwargs)
        o_init = check_class.__init__

        def __init__(self, *args, **kwargs):
            for name, value in kwargs.items():
                # getting field type
                ft = check_class.__annotations__.get(name, None)
                if is_dataclass(ft) and isinstance(value, dict):
                    obj = ft(**value)
                    kwargs[name] = obj
                # Wouldn't be necessary if type annotations as strings in
                # dataclasses were valid…
                elif isinstance(ft, str) and isinstance(value, dict):
                    # Escape - to make it possible to translate payload keys in
                    # dataclass field.
                    new_dict = {}
                    for k, v in value.items():
                        new_dict[k.replace("-", "_")] = v
                    value = new_dict
                    obj = getattr(signald_api, ft)(**value)
                    kwargs[name] = obj
                # Even more tricky, List of dataclass, did not find a better
                # way
                elif str(ft).startswith("typing.List[ForwardRef"):
                    ft = str(ft).split("'")[1]
                    obj = [getattr(signald_api, ft)(**x) for x in value]
                    kwargs[name] = obj

                o_init(self, *args, **kwargs)

        check_class.__init__ = __init__

        return check_class

    return wrapper(args[0]) if args else wrapper


def locals_to_request(d: dict):
    """
    Helper for the generated bindings.
    """
    request = {}
    for k, v in d.items():
        if v is None or k == "self":
            continue
        k = k.replace("-", "_")
        if is_dataclass(v):
            request[k] = asdict_non_none(v)
        else:
            request[k] = v
    return request


def asdict_non_none(obj, *, dict_factory=dict):
    """
    Taken from python standard library, modified to handle our nested dataclasses
    and avoid None values
    """
    if not is_dataclass(obj):
        raise TypeError("asdict() should be called on dataclass instances")
    return _asdict_inner(obj, dict_factory)


def _asdict_inner(obj, dict_factory):
    """
    Taken from python standard library, modified to handle our nested dataclasses
    """
    if is_dataclass(obj):
        result = []
        for f in fields(obj):
            value = _asdict_inner(getattr(obj, f.name), dict_factory)
            if value is None:
                continue
            result.append((f.name, value))
        return dict_factory(result)
    elif isinstance(obj, tuple) and hasattr(obj, "_fields"):
        return type(obj)(*[_asdict_inner(v, dict_factory) for v in obj])
    elif isinstance(obj, (list, tuple)):
        return type(obj)(_asdict_inner(v, dict_factory) for v in obj)
    elif isinstance(obj, dict):
        return type(obj)(
            (_asdict_inner(k, dict_factory), _asdict_inner(v, dict_factory))
            for k, v in obj.items()
        )
    else:
        return copy.deepcopy(obj)


_FIELDS = "__dataclass_fields__"



import asyncio
import typing

from pysignald_async.error import SignaldException
from pysignald_async.generated import *
import pysignald_async.util as util


class SignaldAPI(SignaldGeneratedAPI):
    async def get_response_and_wait_for(
        self, request: dict, validator: typing.Callable
    ):
        future = self.get_future_for(validator)
        try:
            await self.get_response(request)
        except SignaldException:
            self.specific_handlers.remove(future.handler)
            future.cancel()
            raise
        else:
            await future

    async def subscribe(self, username: str):
        """
        Starts receiving messages for the account identified by the argument
        username (a phone number).
        """
        await self.get_response_and_wait_for(
            request={"type": "subscribe", "username": username},
            validator=lambda response: response.get("type") == "listen_started"
            and response.get("data") == username,
        )

    async def unsubscribe(self, username: str):
        """
        Stops receiving message for an phone 'username'.
        """
        await self.get_response_and_wait_for(
            request={"type": "unsubscribe", "username": username},
            validator=lambda response: response.get("type") == "listen_stopped"
            and response.get("data") == username,
        )

    async def link(
        self, deviceName: str, username: str
    ) -> typing.Tuple[str, asyncio.Future]:
        """
        Link signald to an existing signal account.

        Return a URI that needs to be converted to a QR code to be scanned with
        the official signal app, and a future which result is set to True once the
        link is successful.
        """
        linking_successful = self.get_future_for(
            lambda response: response.get("type") == "linking_successful"
            and response.get("data", dict()).get("username") == username
        )
        try:
            response = await self.get_response(
                {"type": "link", "deviceName": deviceName}
            )
        except SignaldException:
            self.specific_handlers.remove(linking_successful.handler)
            linking_successful.cancel()
            raise
        else:
            uri = response.get("data", dict()).get("uri")
            return uri, linking_successful

    async def verify(self, username: str, code: str):
        await self.get_response_and_wait_for(
            request={"type": "verify", "username": username, "code": code},
            validator=lambda response: response.get("type") == "verified"
            and response.get("data", {}).get("username") == username,
        )

    async def register(self, username: str):
        """
        Register signald as the primary signal device for a phone number.
        To complete to process, the SignaldAPI.verify coroutine must then be
        awaited with the code received by SMS.
        """
        await self.get_response_and_wait_for(
            request={"type": "register", "username": username},
            validator=lambda response: response.get("type") == "verification_required"
            and response.get("data", {}).get("username") == username,
        )

    def handle_message(self, payload):
        envelope = JsonMessageEnvelopev1(**payload.get("data", dict()))
        print(envelope)
